
import sympy
import numbers

from .taylornum import taylornum
from .taylorsym import taylorsym


# Check if f has some element that if symbolic in nature.
# If so, we will let the symbolic side handle it
def is_f_symbolic(f):
    if isinstance(f, (list, tuple)):
        return any(isinstance(fi, tuple(sympy.core.all_classes)) for fi in f)
    else:
        return isinstance(f, tuple(sympy.core.all_classes))


# Function to check if this is a sympy symbol or a number
def is_symbol_or_num(fi):
    return isinstance(fi, (tuple(sympy.core.all_classes), numbers.Number))


# Calculate the taylor series.
# This is the main routing function that determines symbolic or
# numerical Taylor series and where to send the data to
def taylor(f, x0=None, n=None):
    # Does f have something symbolic in it
    fsym = is_f_symbolic(f)

    # If n is None, then we know that this is a numeric
    if n is None:
        if x0 is None:
            if fsym:
                return taylorsym(f)
            else:
                # We need to return a lambda function that takes in
                # dt and calculates the values
                return lambda dt: taylornum(f, dt)
        else:
            if fsym:
                return taylorsym(f, x0)
            else:
                return taylornum(f, x0)
    else:
        if x0 is None:
            if fsym:
                return taylorsym(f, n=n)
            else:
                # We need to return a lambda function that takes in
                # dt and calculates the values
                return lambda dt: taylornum(f, n=n)

    # Symbolic. Check that everything is a symbol
    if isinstance(f, list) or isinstance(f, tuple):
        if all(is_symbol_or_num(fi) for fi in f):
            if isinstance(x0, list) or isinstance(x0, tuple):
                if len(f) == len(x0):
                    return taylorsym(f, x0, n)
                else:
                    error_msg = 'ERROR: Size of f and x0 do not match'
                    raise RuntimeError(error_msg)
            else:
                error_msg = 'x0 is not the same size as f'
                raise RuntimeError(error_msg)
        else:
            emsg = 'ERROR: Expected list to be all symbolic expressions.'
            raise RuntimeError(emsg)
    elif isinstance(f, tuple(sympy.core.all_classes)) \
            or isinstance(f, numbers.Number):
        return taylorsym(f, x0, n)
    else:
        raise RuntimeError('ERROR: Expected symbolic expressions.')




import sympy
import numbers
import itertools

from ohtaylor.taylorutils import ktuples_nsum, factorial_tuple, \
    create_symbol_point_list, sort_symbols
from ohtaylor.diffcalc import DiffCalcCached


# Find the Taylor series when f:R^1 -> R^1
def taylorsym_1_1(x, f, x0):
    # If the input is given as a list or tuple of one variable,
    # change it to be a number
    if isinstance(x0, (list, tuple)):
        x0 = x0[0]

    # Initialize the iteration
    fi_1x = f
    fix = f

    for i in itertools.count():
        # Calculate the Taylor series term
        px_i = fix.subs(x, x0) / sympy.factorial(i) * (x - x0)**i
        yield px_i

        # Calculate the next derivative
        fix = sympy.diff(fi_1x, x)
        fi_1x = fix


# Find the Taylor series for f:R^N -> R
def taylorsym_N_1(x, f, x0):
    # Number of input variables
    k = len(x)
    vars = create_symbol_point_list(x, x0)
    d = DiffCalcCached()
    syms, ivs = zip(*vars)

    # Generate the Taylor series
    for i in itertools.count():
        tuples = ktuples_nsum(k, i)
        px_i = 0
        for t in tuples:
            fix = d.diff(f, syms, t)
            fix = fix.subs(vars)
            for j in range(0, k):
                fix *= (syms[j] - ivs[j])**t[j]
            px_i += fix / factorial_tuple(t)
        yield px_i


# Send to R^1 or R^n
def taylorsym_d(f, x0, n):
    if isinstance(f, numbers.Number):
        return f

    # Get the independent variables
    symbols = f.free_symbols

    if x0 is None:
        x0_syms = []
        for symbol in sort_symbols(symbols):
            x0_syms.append(sympy.Symbol(symbol.name + '0'))
        x0 = x0_syms

    # Dispatch to the right place
    if len(symbols) == 1:
        [x] = symbols
        gen = taylorsym_1_1(x, f, x0)
    else:
        gen = taylorsym_N_1(symbols, f, x0)

    # User wants the generator
    if n is None:
        return gen

    # Use the generator to create the requested polynomial
    px = 0
    for i in range(0, n+1):
        px += next(gen)
    return px


# Send each of the dimensions of R^M output to be evaluated independently
def taylorsym_l(f, x0, n):
    px = [None]*len(f)
    for i in range(0, len(f)):
        if x0 is None:
            px[i] = taylorsym_d(f[i], x0, n=n)
        else:
            px[i] = taylorsym_d(f[i], x0[i], n)
    return px


# Figure out the dimensions and dispatch to the right Taylor series evaluation
# Check that the inputs are valid and match up
def taylorsym(f, x0=None, n=None):
    if isinstance(f, list) or isinstance(f, tuple):
        return taylorsym_l(f, x0, n)
    else:
        return taylorsym_d(f, x0, n)




import unittest

from testnum import TestNumericalTaylor  # noqa
from testsym import TestSymbolicTaylor   # noqa
from testutils import TestTaylorUtils    # noqa
from testdiffcalc import TestDiffCalc    # noqa


def main():
    unittest.main()


if __name__ == "__main__":
    main()




import os
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


short_description = \
    'Multi-Variate Taylor Series Approximation/Calculation Library for Python'

setup(
    name='ohtaylor',
    version='1.1.0',
    packages=['ohtaylor'],
    url='https://gitlab.com/mochan/ohtaylor',
    license='Apache 2.0',
    author='Mochan Shrestha',
    author_email='',
    description=short_description,
    long_description=read('README'),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
)
